==> kite-1.0b6.tar.gz: kite-1.0b6/modules/__internal/_cinvoke_structure.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_construct) {
    CInvContext *ctx;
    CInvStructure *st;
    kite_object_t *ctxObj;
    
    (void)thd;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    
    st = cinv_structure_create(ctx);
    
    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)st;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_destruct) {
    KITE_NO_ARGS;
    (void)thd;
    
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    cinv_structure_delete(ctx, st);
}

/*****************************************************************************
 * Add simple member to structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_addValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *nameStr, *typeInt;
    char *name;
    int type;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(nameStr, 1);
    KITE_GET_METHOD_ARGUMENT(typeInt, 2);

    name = KITE_GET_STRING_VALUE(nameStr);
    type = KITE_GET_INTEGER(typeInt);
    
    cinv_structure_addmember_value(ctx, st, name, type);
}

/*****************************************************************************
 * Add structure type to structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_addStruct) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    CInvStructure *st_struct;
    kite_object_t *nameStr, *typeStruct;
    char *name;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(nameStr, 1);
    KITE_GET_METHOD_ARGUMENT(typeStruct, 2);

    name = KITE_GET_STRING_VALUE(nameStr);
    st_struct = (CInvStructure*)typeStruct->builtin_data.listvalue.cdr;
    
    cinv_structure_addmember_struct(ctx, st, name, st_struct);
}

/*****************************************************************************
 * Get structure's size.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_getSize) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    int sz;
    
    KITE_NO_ARGS;
    (void)thd;
    cinv_structure_getsize(ctx, st, &sz);
    kite_vm_return(thd, kite_new_integer(thd, sz));
}

/*****************************************************************************
 * Finish structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_finish) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    (void)thd;
    cinv_structure_finish(ctx, st);
}

/*****************************************************************************
 * Create new instance.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_newInstance) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    kite_vm_return(thd, 
        kite_new_integer(thd, 
            (long)cinv_structure_create_instance(ctx, st)
        )
    );
}

/*****************************************************************************
 * Destroy instance.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_destroyInstance) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    void *instance;
    kite_object_t *instInt;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(instInt, 1);

    instance = (void*)KITE_GET_INTEGER(instInt);
    cinv_structure_delete_instance(ctx, instance);
}

/*****************************************************************************
 * Set instance value.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_setInstanceValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *instInt, *nameStr, *val, *typeStr;
    char *name, *param_type_string;
    void *instance, *param_buf;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(instInt, 1);
    KITE_GET_METHOD_ARGUMENT(nameStr, 2);
    KITE_GET_METHOD_ARGUMENT(val, 3);
    KITE_GET_METHOD_ARGUMENT(typeStr, 4);
    
    name = KITE_GET_STRING_VALUE(nameStr);
    instance = (void*)KITE_GET_INTEGER(instInt);
    param_type_string = KITE_GET_STRING_VALUE(typeStr);
    switch(*param_type_string)
    {
        case 'c':
        {
            param_buf = malloc(sizeof(char));
            *(char*)(param_buf) = (char)KITE_GET_INTEGER(val);
            break;
        }
        case 's':
        {
            param_buf = malloc(sizeof(short));
            *(short*)(param_buf) = (short)KITE_GET_INTEGER(val);
            break;
        }
        case 'i':
        {
            param_buf = malloc(sizeof(int));
            *(int*)(param_buf) = (int)KITE_GET_INTEGER(val);
            break;
        }
        case 'l':
        {
            param_buf = malloc(sizeof(long));
            *(long*)(param_buf) = (long)KITE_GET_INTEGER(val);
            break;
        }
        case 'p':
        {
            param_buf = malloc(sizeof(int));
            *(int*)(param_buf) = (int)KITE_GET_INTEGER(val);
            break;
        }
        case 't':
        {
            param_buf = malloc(sizeof(int));
            *(char**)(param_buf) = KITE_GET_STRING_VALUE(val);
            break;
        }
        case '2':
        {
            param_buf = malloc(sizeof(cinv_int16_t));
            *(cinv_int16_t*)(param_buf) = (cinv_int16_t)KITE_GET_INTEGER(val);
            break;
        }
        case '4':
        {
            param_buf = malloc(sizeof(cinv_int32_t));
            *(cinv_int32_t*)(param_buf) = (cinv_int32_t)KITE_GET_INTEGER(val);
            break;
        }
        case '8':
        {
            param_buf = malloc(sizeof(cinv_int64_t));
            *(cinv_int64_t*)(param_buf) = (cinv_int64_t)KITE_GET_INTEGER(val);
            break;
        }
        case 'e':
        {
            param_buf = malloc(sizeof(long long));
            *(long long*)(param_buf) = (long long)KITE_GET_INTEGER(val);
            break;
        }
        case 'f':
        {
            param_buf = malloc(sizeof(float));
            *(float*)(param_buf) = (float)KITE_GET_FLOAT(val);
            break;
        }
        case 'd':
        {
            param_buf = malloc(sizeof(double));
            *(double*)(param_buf) = (double)KITE_GET_FLOAT(val);
            break;
        }
        case 'r':
        {
            param_buf = (void*)KITE_GET_INTEGER(val);
        }
        default:
        {
            assert(0);
        }
    }

    cinv_structure_instance_setvalue(ctx, st, instance, name, param_buf);
    if (*param_type_string != 'r') free(param_buf);
}

/*****************************************************************************
 * Get instance value.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_getInstanceValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *instInt, *nameStr, *typeStr, *tmp;
    char *name, *param_type_string;
    void *instance, *ret_buf;
    
    KITE_GET_METHOD_ARGUMENT(instInt, 1);
    KITE_GET_METHOD_ARGUMENT(nameStr, 2);
    KITE_GET_METHOD_ARGUMENT(typeStr, 3);
    
    name = KITE_GET_STRING_VALUE(nameStr);
    instance = (void*)KITE_GET_INTEGER(instInt);
    param_type_string = KITE_GET_STRING_VALUE(typeStr);
    
    ret_buf = cinv_structure_instance_getvalue(ctx, st, instance, name);
    
    if (ret_buf)
    {
        switch(*param_type_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)ret_buf);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)ret_buf);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)ret_buf);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, *(long long*)ret_buf);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)ret_buf);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)ret_buf);
                break;
            }
            case 'r':
            case 'p':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 't':
            {
                tmp = kite_new_string(thd, *(char**)ret_buf);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)ret_buf);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)ret_buf);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, *(cinv_int64_t*)ret_buf);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        kite_vm_push(thd, tmp);
    }
    else
    {
        kite_vm_push(thd, kite_new_null(thd));
    }
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_structure) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_structure");
    kite_set_docstring(newclass, "CInvokeStructure internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_construct, 
                                  "Class constructor.", 1,
                                  "ctx", "CInvokeContext instance"));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "add_value",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_addValue, 
                                  "Add simple type to structure definition", 2,
                                  "name", "Name of member.",
                                  "type", "Member's type."));
    kite_add_method(thread, newclass, "add_struct",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_addStruct, 
                                  "Add structure type to structure definition", 2,
                                  "name", "Name of member.",
                                  "type", "Member's type."));
    kite_add_method(thread, newclass, "size",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_getSize, 
                                  "Get structure's size.", 0));
    kite_add_method(thread, newclass, "finish",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_finish, 
                                  "Finish structure definition.", 0));
    kite_add_method(thread, newclass, "new_instance",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_newInstance, 
                                  "Create new instance of structure.", 0));
    kite_add_method(thread, newclass, "destroy_instance",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_destroyInstance, 
                                  "Destroy instance of structure.", 1,
                                  "inst", "Instance of structure to destroy."));
    kite_add_method(thread, newclass, "set_instance_value",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_setInstanceValue, 
                                  "Set value of given member.", 4,
                                  "inst", "Instance to set value for.",
                                  "name", "Name of member to set.",
                                  "value", "Value to set to.",
                                  "type", "Type of given member."));
    kite_add_method(thread, newclass, "get_instance_value",
        kite_new_method_with_docs(thread, FUNC_COMPILED, CInvokeStructure_getInstanceValue, 
                                  "Get value of given member.", 3,
                                  "inst", "Instance to retrieve value from.",
                                  "name", "Name of member to retrieve.",
                                  "type", "Type of given member."));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_structure", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_structure", newclass);
}

==> kite_1.0.2.orig.tar.gz: kite-1.0.2/modules/__internal/_cinvoke_structure.c <==
/*****************************************************************************
 * Copyright (c) 2008, Mooneer Salem
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Kite Language organization nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY MOONEER SALEM ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL MOONEER SALEM BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ****************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "kite_object.h"
#include "vm/kite_vm.h"
#include "config.h"

#ifdef BUILD_CINVOKE
#include "cinvoke/cinvoke.h"

/*****************************************************************************
 * Class constructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_construct) {
    CInvContext *ctx;
    CInvStructure *st;
    kite_object_t *ctxObj;
    
    (void)thd;
    
    KITE_GET_METHOD_ARGUMENT(ctxObj, 1);
    
    ctx = (CInvContext*)ctxObj->builtin_data.filevalue;
    
    st = cinv_structure_create(ctx);
    
    this->builtin_data.listvalue.car = 
        (kite_object_t*)ctx;
    this->builtin_data.listvalue.cdr =
        (kite_object_t*)st;
}

/*****************************************************************************
 * Class destructor
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_destruct) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    (void)thd;
    
    cinv_structure_delete(ctx, st);
}

/*****************************************************************************
 * Add simple member to structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_addValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *nameStr, *typeInt;
    char *name;
    int type;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(nameStr, 1);
    KITE_GET_METHOD_ARGUMENT(typeInt, 2);

    name = KITE_GET_STRING_VALUE(nameStr);
    type = KITE_GET_INTEGER(typeInt);
    
    cinv_structure_addmember_value(ctx, st, name, type);
}

/*****************************************************************************
 * Add structure type to structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_addStruct) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    CInvStructure *st_struct;
    kite_object_t *nameStr, *typeStruct;
    char *name;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(nameStr, 1);
    KITE_GET_METHOD_ARGUMENT(typeStruct, 2);

    name = KITE_GET_STRING_VALUE(nameStr);
    st_struct = (CInvStructure*)typeStruct->builtin_data.listvalue.cdr;
    
    cinv_structure_addmember_struct(ctx, st, name, st_struct);
}

/*****************************************************************************
 * Get structure's size.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_getSize) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    int sz;
    
    KITE_NO_ARGS;
    (void)thd;
    cinv_structure_getsize(ctx, st, &sz);
    kite_vm_return(thd, kite_new_integer(thd, sz));
}

/*****************************************************************************
 * Finish structure.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_finish) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    (void)thd;
    cinv_structure_finish(ctx, st);
}

/*****************************************************************************
 * Create new instance.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_newInstance) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    
    KITE_NO_ARGS;
    kite_vm_return(thd, 
        kite_new_integer(thd, 
            (long)cinv_structure_create_instance(ctx, st)
        )
    );
}

/*****************************************************************************
 * Destroy instance.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_destroyInstance) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    void *instance;
    kite_object_t *instInt;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(instInt, 1);

    instance = (void*)KITE_GET_INTEGER(instInt);
    cinv_structure_delete_instance(ctx, instance);
}

/*****************************************************************************
 * Set instance value.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_setInstanceValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *instInt, *nameStr, *val, *typeStr;
    char *name, *param_type_string;
    void *instance, *param_buf;
    
    (void)thd;
    KITE_GET_METHOD_ARGUMENT(instInt, 1);
    KITE_GET_METHOD_ARGUMENT(nameStr, 2);
    KITE_GET_METHOD_ARGUMENT(val, 3);
    KITE_GET_METHOD_ARGUMENT(typeStr, 4);
    
    name = KITE_GET_STRING_VALUE(nameStr);
    instance = (void*)KITE_GET_INTEGER(instInt);
    param_type_string = KITE_GET_STRING_VALUE(typeStr);
    switch(*param_type_string)
    {
        case 'c':
        {
            param_buf = malloc(sizeof(char));
            *(char*)(param_buf) = (char)KITE_GET_INTEGER(val);
            break;
        }
        case 's':
        {
            param_buf = malloc(sizeof(short));
            *(short*)(param_buf) = (short)KITE_GET_INTEGER(val);
            break;
        }
        case 'i':
        {
            param_buf = malloc(sizeof(int));
            *(int*)(param_buf) = (int)KITE_GET_INTEGER(val);
            break;
        }
        case 'l':
        {
            param_buf = malloc(sizeof(long));
            *(long*)(param_buf) = (long)KITE_GET_INTEGER(val);
            break;
        }
        case 'p':
        {
            param_buf = malloc(sizeof(int));
            *(int*)(param_buf) = (int)KITE_GET_INTEGER(val);
            break;
        }
        case 't':
        {
            param_buf = malloc(sizeof(int));
            *(char**)(param_buf) = KITE_GET_STRING_VALUE(val);
            break;
        }
        case '2':
        {
            param_buf = malloc(sizeof(cinv_int16_t));
            *(cinv_int16_t*)(param_buf) = (cinv_int16_t)KITE_GET_INTEGER(val);
            break;
        }
        case '4':
        {
            param_buf = malloc(sizeof(cinv_int32_t));
            *(cinv_int32_t*)(param_buf) = (cinv_int32_t)KITE_GET_INTEGER(val);
            break;
        }
        case '8':
        {
            param_buf = malloc(sizeof(cinv_int64_t));
            *(cinv_int64_t*)(param_buf) = (cinv_int64_t)KITE_GET_INTEGER(val);
            break;
        }
        case 'e':
        {
            param_buf = malloc(sizeof(long long));
            *(long long*)(param_buf) = (long long)KITE_GET_INTEGER(val);
            break;
        }
        case 'f':
        {
            param_buf = malloc(sizeof(float));
            *(float*)(param_buf) = (float)KITE_GET_FLOAT(val);
            break;
        }
        case 'd':
        {
            param_buf = malloc(sizeof(double));
            *(double*)(param_buf) = (double)KITE_GET_FLOAT(val);
            break;
        }
        case 'r':
        {
            param_buf = (void*)KITE_GET_INTEGER(val);
        }
        default:
        {
            assert(0);
        }
    }

    cinv_structure_instance_setvalue(ctx, st, instance, name, param_buf);
    if (*param_type_string != 'r') free(param_buf);
}

/*****************************************************************************
 * Get instance value.
 ****************************************************************************/
KITE_CLASS_METHOD(CInvokeStructure_getInstanceValue) {
    CInvContext *ctx = (CInvContext*)this->builtin_data.listvalue.car;
    CInvStructure *st = (CInvStructure*)this->builtin_data.listvalue.cdr;
    kite_object_t *instInt, *nameStr, *typeStr, *tmp;
    char *name, *param_type_string;
    void *instance, *ret_buf;
    
    KITE_GET_METHOD_ARGUMENT(instInt, 1);
    KITE_GET_METHOD_ARGUMENT(nameStr, 2);
    KITE_GET_METHOD_ARGUMENT(typeStr, 3);
    
    name = KITE_GET_STRING_VALUE(nameStr);
    instance = (void*)KITE_GET_INTEGER(instInt);
    param_type_string = KITE_GET_STRING_VALUE(typeStr);
    
    ret_buf = cinv_structure_instance_getvalue(ctx, st, instance, name);
    
    if (ret_buf)
    {
        switch(*param_type_string)
        {
            case 'c':
            {
                tmp = kite_new_integer(thd, *(char*)ret_buf);
                break;
            }
            case 's':
            {
                tmp = kite_new_integer(thd, *(short*)ret_buf);
                break;
            }
            case 'i':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 'l':
            {
                tmp = kite_new_integer(thd, *(long*)ret_buf);
                break;
            }
            case 'e':
            {
                tmp = kite_new_integer(thd, (long)*(long long*)ret_buf);
                break;
            }
            case 'f':
            {
                tmp = kite_new_float(thd, *(float*)ret_buf);
                break;
            }
            case 'd':
            {
                tmp = kite_new_float(thd, *(double*)ret_buf);
                break;
            }
            case 'r':
            case 'p':
            {
                tmp = kite_new_integer(thd, *(int*)ret_buf);
                break;
            }
            case 't':
            {
                tmp = kite_new_string(thd, *(char**)ret_buf);
                break;
            }
            case '2':
            {
                tmp = kite_new_integer(thd, *(cinv_int16_t*)ret_buf);
                break;
            }
            case '4':
            {
                tmp = kite_new_integer(thd, *(cinv_int32_t*)ret_buf);
                break;
            }
            case '8':
            {
                tmp = kite_new_integer(thd, (long)*(cinv_int64_t*)ret_buf);
                break;
            }
            default:
            {
                assert(0);
            }
        }

        kite_vm_push(thd, tmp);
    }
    else
    {
        kite_vm_push(thd, kite_new_null(thd));
    }
}
#endif

KITE_MODULE_INITIALIZER(_cinvoke_structure) {
    /*
     * Create new class object. inherits is typically NULL for packages
     * and System.object.
     */
    kite_object_t *inherits = 
        kite_dereference_and_load(thread, "System.object");
    kite_object_t *newclass = 
        kite_new_class(thread, inherits, "_cinvoke_structure");
    kite_set_docstring(newclass, "CInvokeStructure internal code.");

#ifdef BUILD_CINVOKE    
    /* Add methods. */
    kite_add_method(thread, newclass, "__construct__",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_construct, 
                                  "Class constructor.", 1,
                                  "ctx", "CInvokeContext instance"));
    kite_add_method(thread, newclass, "destruct",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_destruct, 
                                  "Free associated memory and resources.", 0));
    kite_add_method(thread, newclass, "add_value",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_addValue, 
                                  "Add simple type to structure definition", 2,
                                  "name", "Name of member.",
                                  "type", "Member's type."));
    kite_add_method(thread, newclass, "add_struct",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_addStruct, 
                                  "Add structure type to structure definition", 2,
                                  "name", "Name of member.",
                                  "type", "Member's type."));
    kite_add_method(thread, newclass, "size",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_getSize, 
                                  "Get structure's size.", 0));
    kite_add_method(thread, newclass, "finish",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_finish, 
                                  "Finish structure definition.", 0));
    kite_add_method(thread, newclass, "new_instance",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_newInstance, 
                                  "Create new instance of structure.", 0));
    kite_add_method(thread, newclass, "destroy_instance",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_destroyInstance, 
                                  "Destroy instance of structure.", 1,
                                  "inst", "Instance of structure to destroy."));
    kite_add_method(thread, newclass, "set_instance_value",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_setInstanceValue, 
                                  "Set value of given member.", 4,
                                  "inst", "Instance to set value for.",
                                  "name", "Name of member to set.",
                                  "value", "Value to set to.",
                                  "type", "Type of given member."));
    kite_add_method(thread, newclass, "get_instance_value",
        kite_new_method_compiled_with_docs(thread, CInvokeStructure_getInstanceValue, 
                                  "Get value of given member.", 3,
                                  "inst", "Instance to retrieve value from.",
                                  "name", "Name of member to retrieve.",
                                  "type", "Type of given member."));
#endif

    /* Add to parent to complete loading. */
    kite_add_property(thread, parent, "_cinvoke_structure", 1, NULL);
    kite_set_property(thread, parent, "_cinvoke_structure", newclass);
}
